'''
    File name: detector_utils.py
    Author: [Qoala DS Team]
    Date created: / /2019
    Date last modified: / /2019
    Python Version: >= 3.5
    qoalai version: v0.4.4
    License: MIT License
    Maintainer: [Mochammad F Rahman]
'''

import tensorflow as tf
from qoalai.transfer_learning.image_recognition import *


class Ensemble(): 
    """Class for ensembling two or more networks
    """    
    def __init__(self, classes,
                       ensemble_network = ['resnet', 'densenet'],
                       input_height = 300,
                       input_width = 300,
                       input_channel = 3,
                       tf_lite_architecture=False):
        """Ensemble Constructor
        
        Arguments:
            classes {[list of str]} -- the class list
        
        Keyword Arguments:
            ensemble_network {list} -- [description] (default: {['resnet', 'densenet']})
            input_height {int} -- [description] (default: {300})
            input_width {int} -- [description] (default: {300})
            input_channel {int} -- [description] (default: {3})
        """        
        self.classes = classes
        self.ensemble_networks = ensemble_network
        self.tf_lite_architecture = tf_lite_architecture
        
        if len(self.ensemble_networks) in [2, 3]:
            #print ("==>> INFO: your networks are {} and {}".format(self.ensemble_networks[0], self.ensemble_networks[1]))

            self.input_height = input_height
            self.input_width = input_width
            self.input_channel = input_channel

            self.imrec = ImageRecognition(classes=self.classes,
                            input_height = self.input_height,
                            input_width = self.input_width, 
                            input_channel = self.input_channel)
            
            if self.tf_lite_architecture: 
                self.input_images = self.imrec.input_placeholder
            else: 
                self.input_string = tf.placeholder(tf.string, shape=[None], name='string_input')
                decode = lambda raw_byte_str: tf.image.resize_images(
                                                tf.cast(
                                                    tf.image.decode_jpeg(raw_byte_str, channels=3, name='decoded_image'),
                                                        tf.uint8), 
                                                [self.input_height, self.input_width])
                self.input_images = tf.map_fn(decode, self.input_string, dtype=tf.float32) / 255.0
                
        
        elif len(self.ensemble_networks) == 1: 
            print ("==>> WARNING: your network is only {}, Ensemble is combining TWO or more networks".format(self.ensemble_networks[0]))
        else: 
            print ("==>> WARNING: your networks are more than 2, This current version ONLY SUPPORTs 2 or 3 models ensemble")

    
    def do_ensemble(self, model_list,
                    model_result_path,
                    is_training=False):
        """[summary]
        
        Arguments:
            model_list {[type]} -- [description]
        
        Keyword Arguments:
            is_training {bool} -- [description] (default: {False})
        """
        network_output = {}
        network_vars = {} 
        savers = {}

        for i in self.ensemble_networks: 
            if i == 'densenet':
                network_output[i], _ = self.imrec.build_densenet_base(self.input_images,
                                                                dropout_rate = 0.20,
                                                                is_training = is_training,
                                                                top_layer_depth = 128) 
                network_vars[i] = tf.global_variables(scope='densenet121')

            elif i == 'resnet': 
                network_output[i], _ = self.imrec.build_resnetv2(self.input_images,
                                                            is_training=is_training,
                                                            top_layer_depth = 128) 
                network_vars[i] = tf.global_variables(scope='resnet_v2_101')

            elif i == 'inception': 
                network_output[i], _ = self.imrec.build_inceptionv4_basenet(self.input_images,
                                                            is_training=is_training,
                                                            final_endpoint='Mixed_7a', # 'Mixed_6a, Mixed_5a, Mixed_7a
                                                            top_layer_depth = 128) 
                network_vars[i] = tf.global_variables(scope='InceptionV4')
        
        self.out = tf.add_n(list(network_output.values())) / float(len(network_output)) #(network_output[self.ensemble_networks[0]] + network_output[self.ensemble_networks[1]]) / 2.
        print ("==>> INFO: Building networks success")
        
        session = tf.Session()
        saver_all = tf.train.Saver()
        for idx, i in enumerate(self.ensemble_networks): 
            savers[i] = tf.train.Saver(var_list=network_vars[i])
            savers[i].restore(sess=session, save_path=model_list[idx])
        print ('==>> Load all wieghts success')

        ############  serving model procedure #################
        builder = tf.saved_model.builder.SavedModelBuilder(model_result_path)

        # Create aliase tensors
        # tensor_info_x: for input tensor
        # tensor_info_y: for output tensor
        if self.tf_lite_architecture: 
            tensor_info_x = tf.saved_model.utils.build_tensor_info(self.input_images)
        else: 
            tensor_info_x = tf.saved_model.utils.build_tensor_info(self.input_string)
        tensor_info_y = tf.saved_model.utils.build_tensor_info(self.out)

        # create prediction signature
        prediction_signature = tf.saved_model.signature_def_utils.build_signature_def(
                                inputs={'input': tensor_info_x},
                                outputs={'output': tensor_info_y},
                                method_name=tf.saved_model.signature_constants.PREDICT_METHOD_NAME)

        # build frozen graph
        legacy_init_op = tf.group(tf.tables_initializer(), name='legacy_init_op')
        builder.add_meta_graph_and_variables(session, [tf.saved_model.tag_constants.SERVING],
                                             signature_def_map={'serving_default':prediction_signature},
                                             legacy_init_op=legacy_init_op)
        builder.save()





import cv2 
import numpy as np 

GRID = 19


def nms_phone(map, position, grid_size=1/GRID):
    """Methode for landmark NMS
    
    Arguments:
        map {numpy} -- the output map
        position {numpy} -- the output position 
    
    Keyword Arguments:
        grid_size {[type]} -- [description] (default: {1/GRID})
    
    Returns:
        [type] -- [description]
    """
    img = (map * 255).astype(np.uint8)
    ret, labels = cv2.connectedComponents(img)
    
    # ----------------------------- #
    # labelling the result          #
    # ----------------------------- #
    label_dict = {}
    for i in range(GRID):
        for j in range(GRID):
            tmp = labels[i, j]
            
            if tmp != 0:
                if tmp not in list(label_dict.keys()):
                    label_dict[tmp] = []

                label_dict[tmp].append([map[i, j], i, j])
    
    # ----------------------------- #
    # find the max value for each lbl
    # ignore the non maxima         #
    # ----------------------------- #
    max_val = {}
    for i in label_dict:
        if i not in max_val:
            max_val[i] = []
        
        biggest_val = 0
        for j in label_dict[i]:
            if biggest_val < j[0]:
                biggest_val = j[0]
                max_val[i] = j
    
    # ----------------------------- #
    # for each maximum point        #
    # find the x and y point        #
    # ----------------------------- #
    result = {}
    for i in max_val:
        ctr_x = (max_val[i][2] + position[max_val[i][1], max_val[i][2], 0]) * grid_size
        ctr_y = (max_val[i][1] + position[max_val[i][1], max_val[i][2], 1]) * grid_size

        ctr_x = max(ctr_x, 0)
        ctr_y = max(ctr_y, 0)
        ctr_x = min(ctr_x, 300)
        ctr_y = min(ctr_y, 300)
        result[i] = (max_val[i][0], ctr_x, ctr_y)
    
    # ----------------------------- #
    # if the number of points > 4   #
    # remove the         #
    # ----------------------------- #
    while(True):
        if len(result.keys()) > 4:
            worst = 1.
            worst_point = None
            for i in result:
                if worst > result[i][0]:
                    worst = result[i][0]
                    worst_point = i
                    
            del result[worst_point]
            
        else:
            break  
    
    # ----------------------------- #
    # sorting the point             #
    # ----------------------------- #
    sorted_result = {}
    
    if len(result.keys()) == 4:
        # ----------------------------- #
        # sorting the point             #
        # ----------------------------- #
        for i in result: 
            ctr_x = result[i][1]
            ctr_y = result[i][2]
            
            gradiens_x = []
            gradiens_y = []
            for j in result: 
                if j != i :
                    ctr_x_new = result[j][1]
                    ctr_y_new = result[j][2] 
                    gradien_y = (ctr_y_new - ctr_y)
                    gradien_x = (ctr_x_new - ctr_x)
                    gradiens_x.append(gradien_x)
                    gradiens_y.append(gradien_y)
            
            gradien_x = np.mean(np.array(gradiens_x))
            gradien_y = np.mean(np.array(gradiens_y))
            
            # ----------------------------- #
            # the top letf point            #
            # ----------------------------- #
            if gradien_x >= 0 and gradien_y >=0: 
                if 1 not in sorted_result: 
                    sorted_result[1] = (result[i][0], ctr_x, ctr_y, gradien_x)
                else: 
                    if gradien_x > sorted_result[1][3]: 
                        sorted_result[2] = sorted_result[1]
                        sorted_result[1] = (result[i][0], ctr_x, ctr_y, gradien_x)
                    else: 
                        sorted_result[2] = (result[i][0], ctr_x, ctr_y, gradien_x)

            # ----------------------------- #
            # the top right point           #
            # ----------------------------- # 
            elif gradien_x < 0 and gradien_y >=0: 
                if 2 not in sorted_result: 
                    sorted_result[2] = (result[i][0], ctr_x, ctr_y, gradien_x)
                else:
                    if gradien_x > sorted_result[2][3]: 
                        sorted_result[1] = (result[i][0], ctr_x, ctr_y, gradien_x)
                    else: 
                        sorted_result[1] = sorted_result[2]
                        sorted_result[2] = (result[i][0], ctr_x, ctr_y, gradien_x)

            # ----------------------------- #
            # the bottom right point        #
            # ----------------------------- #    
            elif gradien_x <= 0 and gradien_y < 0: 
                if 4 not in sorted_result: 
                    sorted_result[4] = (result[i][0], ctr_x, ctr_y, gradien_x)
                else: 
                    if gradien_x > sorted_result[4][3]: 
                        sorted_result[3] = (result[i][0], ctr_x, ctr_y, gradien_x)
                    else:
                        sorted_result[3] = sorted_result[4]
                        sorted_result[4] = (result[i][0], ctr_x, ctr_y, gradien_x)

            # ----------------------------- #
            # the bottom left point         #
            # ----------------------------- #
            else:
                if 3 not in sorted_result: 
                    sorted_result[3] = (result[i][0], ctr_x, ctr_y, gradien_x)
                else: 
                    if gradien_x > sorted_result[3][3]: 
                        sorted_result[4] = sorted_result[3]
                        sorted_result[3] = (result[i][0], ctr_x, ctr_y, gradien_x) 
                    else: 
                        sorted_result[4] = (result[i][0], ctr_x, ctr_y, gradien_x) 
                    
    return sorted_result


def draw_res(img, point_dict):
    h, w, c = img.shape
    
    success_status = False
    # ----------------------------- #
    # if the detected point == 4    #
    # get the actual center point   #
    # do perspective transform      #
    # ----------------------------- # 
    if len(list(point_dict.keys())) == 4:

        rect = []
        for idx, i in enumerate(point_dict):
            #ctr_x = int(point_dict[idx+1][1] * w )
            
            if i == 1 or i == 3: 
                ctr_x = max(int(point_dict[idx+1][1] * w) -5, 0)
            else: 
                ctr_x = min(int(point_dict[idx+1][1] * w) +5, w)

            if i == 1 or i == 2: 
                ctr_y = max(int(point_dict[idx+1][2] * h) -5, 0)
            else: 
                ctr_y = min(int(point_dict[idx+1][2] * h) +5, h)

            rect.append([ctr_x, ctr_y])
        
        dst = np.array([[0, 0], [w-1, 0], [0, h-1], [w-1, h-1]], dtype = "float32")
        rect = np.array(rect, dtype='float32')
        M = cv2.getPerspectiveTransform(rect, dst)
        img = cv2.warpPerspective(img, M, (w, h))
        success_status = True

    bg = np.zeros((340, 340, 3))
    bg[20:-20, 20:-20, :] = img
    img = bg
    
    return img, success_status


'''
    File name: yolo.py
    Author: [Mochammad F Rahman, Agil Haykal]
    Date created: / /2019
    Date last modified: 17/08/2019
    Python Version: >= 3.5
    qoalai version: v0.4.4
    License: MIT License
    Maintainer: [Mochammad F Rahman, Agil Haykal]
'''

import json
import random
import tensorflow as tf
from qoalai.tensor_operations import *
from qoalai.object_detector.detector_utils_b import *
from comdutils.file_utils import *


# =============================================== #
# This class is the child of ObjectDetector class #
# in simple_tensor.object_detector.detector_utils #
# =============================================== #
class Yolo(ObjectDetector):
    def __init__(self, 
                 num_of_class,
                 input_height=416, 
                 input_width=416, 
                 grid_height1=32, 
                 grid_width1=32, 
                 grid_height2=16, 
                 grid_width2=16, 
                 grid_height3=8, 
                 grid_width3=8,
                 objectness_loss_alpha=2., 
                 noobjectness_loss_alpha=1., 
                 center_loss_alpha=1., 
                 size_loss_alpha=1., 
                 class_loss_alpha=1.,
                 add_modsig_toshape=False,
                 anchor = [(10, 13), (16, 30), (33, 23), (30, 61), (62, 45), (59, 119), (116, 90), (156, 198), (373, 326)],
                 dropout_rate = 0.8,
                 leaky_relu_alpha = 0.1,
                 convert_to_tflite=False,
                 is_multilabel = False,
                 multilabel_dict = None):
        """[summary]
        
        Arguments:
            label_folder_path {[type]} -- [description]
            dataset_folder_path {[type]} -- [description]
        
        Keyword Arguments:
            input_height {int} -- [description] (default: {512})
            input_width {int} -- [description] (default: {512})
            grid_height {int} -- [description] (default: {128})
            grid_width {int} -- [description] (default: {128})
            output_depth {int} -- [description] (default: {5})
            objectness_loss_alpha {[type]} -- [description] (default: {1.})
            noobjectness_loss_alpha {[type]} -- [description] (default: {1.})
            center_loss_alpha {[type]} -- [description] (default: {0.})
            size_loss_alpha {[type]} -- [description] (default: {0.})
            class_loss_alpha {[type]} -- [description] (default: {0.})
        """

        super(Yolo, self).__init__(num_of_class=num_of_class,
                                        input_height=input_height, 
                                        input_width=input_width, 
                                        grid_height1=grid_height1, 
                                        grid_width1=grid_width1, 
                                        grid_height2=grid_height2, 
                                        grid_width2=grid_width2, 
                                        grid_height3=grid_height3, 
                                        grid_width3=grid_height3,
                                        objectness_loss_alpha=objectness_loss_alpha, 
                                        noobjectness_loss_alpha=noobjectness_loss_alpha, 
                                        center_loss_alpha=center_loss_alpha, 
                                        size_loss_alpha=size_loss_alpha, 
                                        class_loss_alpha=class_loss_alpha,
                                        add_modsig_toshape=add_modsig_toshape,
                                        anchor = anchor,
                                        dropout_rate = dropout_rate,
                                        leaky_relu_alpha = leaky_relu_alpha,
                                        convert_to_tflite=convert_to_tflite,
                                        is_multilabel=is_multilabel,
                                        multilabel_dict=multilabel_dict)

        self.input_placeholder = tf.placeholder(tf.float32, shape=(None, self.input_height, self.input_width, 3))
        self.lr_placeholder = tf.placeholder(tf.float32, shape=[])
        if self.is_multilabel: 
            self.output_placeholder1 = tf.placeholder(tf.float32, shape=(None, 13, 13, 3*(5 + num_of_class + sum(multilabel_dict.values()) )))
            self.output_placeholder2 = tf.placeholder(tf.float32, shape=(None, 26, 26, 3*(5 + num_of_class + sum(multilabel_dict.values()) )))
            self.output_placeholder3 = tf.placeholder(tf.float32, shape=(None, 52, 52, 3*(5 + num_of_class + sum(multilabel_dict.values()) )))
        else: 
            self.output_placeholder1 = tf.placeholder(tf.float32, shape=(None, 13, 13, 3*(5 + num_of_class)))
            self.output_placeholder2 = tf.placeholder(tf.float32, shape=(None, 26, 26, 3*(5 + num_of_class)))
            self.output_placeholder3 = tf.placeholder(tf.float32, shape=(None, 52, 52, 3*(5 + num_of_class)))

        self.optimizer = None
        self.session = None
        self.saver_partial = None
        self.saver_all = None

        self.train_losses = []
        self.o_losses = []
        self.no_losses = []
        self.ct_losses = []
        self.sz_losses = []
        self.cls_losses = []


    def read_target(self, file_path):
        """[summary]
        
        Arguments:
            file_path {[type]} -- [description]
        
        Returns:
            [type] -- [description]
        """
        target = self.read_yolo_labels(file_path)
        return target


    def build_net(self, input_tensor, 
                  network_type='big', 
                  is_training=False):
        """[summary]
        
        Arguments:
            input_tensor {[type]} -- [description]
        
        Keyword Arguments:
            network_type {str} -- [description] (default: {'big'})
            is_training {bool} -- [description] (default: {False})
        """
        with tf.variable_scope('yolo_v3_model'):
            self.build_yolov3_net(inputs=input_tensor, 
                                    network_type=network_type, 
                                    is_training=is_training)


    def batch_generator(self, batch_size, dataset_path, message):
        """Train Generator
        
        Arguments:
            batch_size {integer} -- the size of the batch
            image_name_list {list of string} -- the list of image name
        """
        label_folder_path = dataset_path + "labels/"
        dataset_folder_path = dataset_path + "images/"
        dataset_file_list = get_filenames(dataset_folder_path)
        random.shuffle(dataset_file_list)
        
        print ("------------------------INFO-------------------")
        print ("Image Folder: " + dataset_folder_path)
        print ("Number of Image: " + str(len(dataset_file_list)))
        print ("-----------------------------------------------")

        # Infinite loop.
        idx = 0
        while True:
            x_batch = []
            y_pred1 = []
            y_pred2 = []
            y_pred3 = []

            for i in range(batch_size):
                if idx >= len(dataset_file_list):
                    random.shuffle(dataset_file_list)
                    print ("==>>> INFO: your " + message +" dataset is reshuffled again", idx)
                    idx = 0
                try:
                    tmp_x = cv2.imread(dataset_folder_path + dataset_file_list[idx])
                    tmp_x = cv2.cvtColor(tmp_x, cv2.COLOR_BGR2RGB)
                    tmp_x = cv2.resize(tmp_x, (self.input_width, self.input_height))
                    tmp_x = tmp_x.astype(np.float32) / 255.
                    tmp_y = self.read_target(label_folder_path + dataset_file_list[idx].split('.j')[0] + ".txt")
                    x_batch.append(tmp_x)
                    y_pred1.append(tmp_y[0])
                    y_pred2.append(tmp_y[1])
                    y_pred3.append(tmp_y[2])
                except Exception as e:
                    print ("---------------------------------------------------------------")
                    print ("WARNING: the " + str(e))
                    print ("---------------------------------------------------------------")

                idx += 1
            yield (np.array(x_batch), [np.array(y_pred1), np.array(y_pred2), np.array(y_pred3)])


    def __draw_rect_wrap(self, bbox, img):
        for i in bbox:
            img = cv2.rectangle(img, (i[0], i[1]), (i[2], i[3]), (0, 255, 255), 1) 
        return img


    def visualize_label(self, labels, image): 
        """[summary]

        Arguments:
            labels {[type]} -- label list (3 labels for yolo)
            image {[type]} -- numpy uint8

        Returns:
            [type] -- [description]
        """        
        images = []

        for idx, i in enumerate(labels): 
            tmps = []

            for j in range(3): 
                base = j * (5+self.num_class)
                tmp = i[:, :, base:base + 5+self.num_class]
                tmps.append(tmp)

            for idy, j in enumerate(tmps): 
                mask = j[:, :, 4:5].astype(np.bool)
                grid_size = self.grid_width[idx]

                centerx = (j[:, :, 0:1] + self.grid_position_mask_onx_np[idx][0, :, :, :]) * grid_size
                centery = (j[:, :, 1:2] + self.grid_position_mask_ony_np[idx][0, :, :, :]) * grid_size
                centerx = centerx[mask].astype(np.int32)
                centery = centery[mask].astype(np.int32)

                if idx == 0: 
                    base = 6
                elif idx == 1: 
                    base = 3
                else: 
                    base = 0
                sizex = np.exp(j[:, :, 2:3][mask]) * self.anchor[base + idy][0]
                sizey = np.exp(j[:, :, 3:4][mask]) * self.anchor[base + idy][1]

                x1 = (centerx - 0.5 * sizex).astype(np.int32).reshape((centerx.shape[0], 1))
                x2 = (centerx + 0.5 * sizex).astype(np.int32).reshape((centerx.shape[0], 1))
                y1 = (centery - 0.5 * sizey).astype(np.int32).reshape((centerx.shape[0], 1))
                y2 = (centery + 0.5 * sizey).astype(np.int32).reshape((centerx.shape[0], 1))
                bbox = np.concatenate([x1, y1, x2, y2], axis=-1)

                sliced_cls = []
                for t in range(self.num_class): 
                    tt = j[:, :, 5+t:5+t+1][mask]
                    tt = np.reshape(tt, (tt.shape[0], 1))
                    sliced_cls.append(tt)
                class_ = np.concatenate(sliced_cls, axis=-1)
                class_ = np.argmax(class_, axis=-1)

                tmp_img = image.copy()
                for a, b, c in zip(centerx, centery, class_): 
                    cv2.circle(tmp_img,(a, b), 5, (0,255,0), -1)
                    cv2.putText(tmp_img, str(c), (a,b), cv2.FONT_HERSHEY_SIMPLEX, 1, (100,255,255))

                tmp_img = self.__draw_rect_wrap(bbox, tmp_img)
                images.append(tmp_img)

        return images


    def optimize(self, subdivisions, 
                 iterations, 
                 best_loss, 
                 train_generator, 
                 val_generator, 
                 save_path, learning_rate = 0.0001):
        """[summary]
        
        Arguments:
            subdivisions {[type]} -- [description]
            iterations {[type]} -- [description]
            best_loss {[type]} -- [description]
            train_generator {[type]} -- [description]
            val_generator {[type]} -- [description]
            save_path {[type]} -- [description]
        """
        best_loss = best_loss
        learning_rate = learning_rate
        #nonnan_count = 0
        #notsaved_time = 0
        
        for i in range(iterations):
            sign = '-'
            tmp_all = [] 
            tmp_obj = [] 
            tmp_noobj = [] 
            tmp_ctr = [] 
            tmp_sz = [] 
            tmp_class = [] 

            if self.is_multilabel: 
                tmp_additional_class = {}
                for val in self.multilabel_dict: 
                    tmp_additional_class[val] = []

            # ---------------------------------- #
            # Training Data                      #
            # ---------------------------------- #
            for j in range (subdivisions):
                x_train, y_train = next(train_generator)
                
                feed_dict = {}
                feed_dict[self.input_placeholder] = x_train
                feed_dict[self.output_placeholder1] = y_train[0]
                feed_dict[self.output_placeholder2] = y_train[1]
                feed_dict[self.output_placeholder3] = y_train[2]
                total, obj, noobj, ctr, size, class_l, iou_avg, obj_acc, noobj_acc, class_acc = self.session.run([self.all_losses, 
                                                            self.objectness_losses, 
                                                            self.noobjectness_losses, 
                                                            self.center_losses, 
                                                            self.size_losses,
                                                            self.class_losses,
                                                            self.iou_avg,
                                                            self.obj_acc_avg,
                                                            self.noobj_acc_avg,
                                                            self.class_acc_avg], feed_dict)

                if math.isnan(obj) or math.isnan(noobj) \
                    or math.isnan(ctr) or math.isnan(size) \
                    or math.isnan(class_l): 
                    print("===============>><<==================") 
                    print ("NAN Loss was found")
                    print ('obj loss: {}, no-obj loss: {}, ctr loss: {}, size loss: {}, class loss: {}'.format(obj, noobj, ctr, size, class_l))
                    print("===============>><<==================") 
                    #if learning_rate >= 1e-6: 
                    #    learning_rate = learning_rate - 0.01 * learning_rate
                    self.saver_all.restore(self.session, save_path)
                """
                else: 
                    nonnan_count += 1
                    if nonnan_count >= 100: 
                        learning_rate = learning_rate + 0.001 * learning_rate
                        nonnan_count = 0
                """
                if self.is_multilabel:
                    additional_ls = self.session.run(self.additional_class_losses, feed_dict)
                    for index, val in enumerate(self.multilabel_dict): 
                        tmp_additional_class[val].append(additional_ls[index])

                    additional_acc = self.session.run(self.additional_class_acc_avg, feed_dict)
                
                feed_dict[self.lr_placeholder] = learning_rate
                self.session.run(self.optimizer, feed_dict=feed_dict)

                tmp_all.append(total)
                tmp_obj.append(obj)
                tmp_noobj.append(noobj)
                tmp_ctr.append(ctr)
                tmp_sz.append(size)
                tmp_class.append(class_l)
                
                if self.is_multilabel: 
                    print ("> Train sub", j, ': iou: ', round(iou_avg*100, 1), 
                            'obj acc: ', round(obj_acc*100, 1), 'noobj acc: ', round(noobj_acc*100, 1), 
                            'main-cls-acc: ', round(class_acc*100, 1), "additional-cls-acc: ", additional_acc, 
                            'lr: ', learning_rate)
                else: 
                    print ("> Train sub", j, ': iou: ', round(iou_avg*100, 1), 
                            'obj acc: ', round(obj_acc*100, 1), 'noobj acc: ', round(noobj_acc*100, 1), 
                            'class acc: ', round(class_acc*100, 1), 'lr: ', learning_rate)
                
            # ------------------------------- #
            # validating the data             #
            # ------------------------------- #
            x_val, y_val = next(val_generator)
            val_feed_dict = {}
            val_feed_dict[self.input_placeholder] = x_val
            val_feed_dict[self.output_placeholder1] = y_val[0]
            val_feed_dict[self.output_placeholder2] = y_val[1]
            val_feed_dict[self.output_placeholder3] = y_val[2]
            total_val, _, _, _, _, _, iou_avg_val, obj_acc_val, noobj_acc_val, class_acc_val = self.session.run([self.all_losses, 
                                                        self.objectness_losses, 
                                                        self.noobjectness_losses, 
                                                        self.center_losses, 
                                                        self.size_losses,
                                                        self.class_losses,
                                                        self.iou_avg,
                                                        self.obj_acc_avg,
                                                        self.noobj_acc_avg,
                                                        self.class_acc_avg], val_feed_dict)
            
            total = sum(tmp_all)/(len(tmp_all) + 0.001)
            obj =  sum(tmp_obj)/(len(tmp_obj) + 0.001)
            noobj = sum(tmp_noobj)/(len(tmp_noobj) + 0.001)
            ctr = sum(tmp_ctr)/(len(tmp_ctr) + 0.001)
            size = sum(tmp_sz)/(len(tmp_sz) + 0.001)
            class_l = sum(tmp_class)/(len(tmp_class) + 0.001)

            if self.is_multilabel: 
                additional_l = {}
                for val in self.multilabel_dict: 
                    additional_l[val] = sum(tmp_additional_class[val])/len(tmp_additional_class[val])
            
            self.train_losses.append(total)
            self.o_losses.append(obj)
            self.no_losses.append(noobj)
            self.ct_losses.append(ctr)
            self.sz_losses.append(size)
            self.cls_losses.append(class_l)
            
            # ------------------------------- #
            # save the model                  #
            # ------------------------------- #
            if best_loss > total_val:
                best_loss = total_val
                sign = "************* model saved"
                self.saver_all.save(self.session, save_path)
                #notsaved_time = 0
            
            print ("> Val epoch :", 'iou: ', round(iou_avg_val*100, 3), 
                    'obj acc: ', round(obj_acc_val*100, 3), 'noobj acc: ', round(noobj_acc_val*100, 3), 
                    'main-cls-acc: ', round(class_acc_val*100, 3))
            
            print ('eph: ', i, 'ttl-loss: ', round(total, 2), 'obj-loss: ', round(obj*100, 2), 
                    'noobj-loss: ', round(noobj*100, 2), 'ctr-loss: ', round(ctr*100, 2), 
                    'size-loss: ', round(size*100, 2),  'main-cls-loss: ', round(class_l*100, 2), sign)
            
            if self.is_multilabel: 
                print ("      additional-cls-loss: ", additional_l)

    
    def check_val_data(self, val_generator):
        """[summary]
        
        Arguments:
            val_generator {[type]} -- [description]
        """
        x_val, y_val = next(val_generator)
        val_feed_dict = {}
        val_feed_dict[self.input_placeholder] = x_val
        val_feed_dict[self.output_placeholder1] = y_val[0]
        val_feed_dict[self.output_placeholder2] = y_val[1]
        val_feed_dict[self.output_placeholder3] = y_val[2]
        total, obj, noobj, ctr, size, class_l, iou_avg_val, obj_acc_val, noobj_acc_val, class_acc_val = self.session.run([self.all_losses, 
                                                    self.objectness_losses, 
                                                    self.noobjectness_losses, 
                                                    self.center_losses, 
                                                    self.size_losses,
                                                    self.class_losses,
                                                    self.iou_avg,
                                                    self.obj_acc_avg,
                                                    self.noobj_acc_avg,
                                                    self.class_acc_avg], val_feed_dict)

        print ("> iou: ", round(iou_avg_val*100, 3), 'obj acc: ', round(obj_acc_val*100, 3), \
                'noobj acc: ', round(noobj_acc_val*100, 3), 'class acc: ', round(class_acc_val*100, 3))

        print ('ttl loss: ', round(total, 2), 'obj loss: ', round(obj*100, 2), \
                'noobj loss: ', round(noobj*100, 2), 'ctr loss: ', round(ctr*100, 2), 'size loss: ', round(size*100, 2),  round(class_l*100, 2))
    


    



'''
    File name: test.py
    Author: [Mochammad F Rahman]
    Date created: / /2019
    Date last modified: 17/07/2019
    Python Version: >= 3.5
    Qoalai version: v0.6.2
    License: MIT License
    Maintainer: [Mochammad F Rahman]
'''

import tensorflow as tf
from tensorflow.python import control_flow_ops


def new_weights(shape, 
                name, 
                data_type=tf.float32):
    """
    Creating new trainable tensor (filter) as weight
    Args:
        shape:		a list of integer as the shape of this weight.
                - example (convolution case), [filter height, filter width, input channels, output channels]
                - example (fully connected case), [num input, num output]
        name:		a string, basic name of this filter/weight
    Return:
        a trainable weight/filter tensor with float 32 data type
    """
    return tf.Variable(tf.random.truncated_normal(shape, stddev=0.05, dtype=data_type), dtype=data_type, name='weight_'+str(name))


def new_biases(length, 
               name, 
               data_type=tf.float32):
    """
    Creating new trainable tensor as bias
    Args:
        length:		an integer, the num of output features 
                - Note, number output neurons = number of bias values
        name:		a string, basic name of this bias
    Return:
        a trainable bias tensor with float 32 data type
    """
    return tf.Variable(tf.constant(0.05, shape=[length], dtype=data_type), dtype=data_type, name='bias_'+str(name))


def new_fc_layer(input, 
                 num_inputs,
                 num_outputs, 
                 name=None, 
                 dropout_val=0.85, 
                 activation="LRELU",
                 lrelu_alpha=0.2, 
                 data_type=tf.float32,
                 is_training=True,
                 use_bias=True): 
    """[summary]
    
    Arguments:
        input {[type]} -- [description]
        num_outputs {[type]} -- [description]
    
    Keyword Arguments:
        name {[type]} -- [description] (default: {None})
        dropout_val {float} -- [description] (default: {0.85})
        activation {str} -- [description] (default: {"LRELU"})
        lrelu_alpha {float} -- [description] (default: {0.2})
        data_type {[type]} -- [description] (default: {tf.float32})
        is_training {bool} -- [description] (default: {True})
        use_bias {bool} -- [description] (default: {True})
    
    Returns:
        [type] -- [description]
    """
    weights = new_weights(shape=[num_inputs, num_outputs], name=name, data_type=data_type)
    layer = tf.matmul(input, weights)

    if use_bias:
        biases = new_biases(length=num_outputs, name=name, data_type=data_type)
        layer += biases

    if activation in ["RELU", "relu", "Relu"]:
        layer = tf.nn.relu(layer)
    elif activation in ["LRELU", "lrelu", "Lrelu"]:
        layer = tf.nn.leaky_relu(layer, alpha=lrelu_alpha)
    elif activation in ["SELU", "selu", "Selu"]:
        layer = tf.nn.selu(layer)
    elif activation in ["ELU", "elu", "Elu"]:
        layer = tf.nn.elu(layer)
    elif activation in ["SIGMOID", "sigmoid", "Sigmoid"]:
        layer = tf.nn.sigmoid(layer)
    elif activation in ["SOFTMAX", "softmax", "Softmax"]:
        layer == tf.nn.softmax(layer)

    if is_training: 
        layer = tf.nn.dropout(layer, dropout_val)
    return layer


def new_conv1d_layer(input, 
                     filter_shape, 
                     name=None, 
                     dropout_val=0.85, 
                     activation='LRELU',
                     lrelu_alpha=0.2,  
                     padding='SAME', 
                     strides=1, 
                     data_type=tf.float32, 
                     is_training=True,
                     use_bias=True,
                     use_batchnorm=False):
    """[summary]
    
    Arguments:
        input {[type]} -- [description]
        filter_shape {[type]} -- [description]
    
    Keyword Arguments:
        name {[type]} -- [description] (default: {None})
        dropout_val {float} -- [description] (default: {0.85})
        activation {str} -- [description] (default: {'LRELU'})
        lrelu_alpha {float} -- [description] (default: {0.2})
        padding {str} -- [description] (default: {'SAME'})
        strides {int} -- [description] (default: {1})
        data_type {[type]} -- [description] (default: {tf.float32})
        is_training {bool} -- [description] (default: {True})
        use_bias {bool} -- [description] (default: {True})
        use_batchnorm {bool} -- [description] (default: {False})
    
    Returns:
        [type] -- [description]
    """
    shape = filter_shape
    weights = new_weights(shape=shape, name=name, data_type=data_type)
    layer = tf.nn.conv1d(input, filters=weights, stride=strides, padding=padding, name='convolution1d_' + str(name))

    if use_bias:
        biases = new_biases(length=filter_shape[2], name=name, data_type=data_type)
        layer += biases

    if use_batchnorm:
        layer = batch_norm(inputs=layer, training=is_training)

    if activation in ["RELU", "relu", "Relu"]:
        layer = tf.nn.relu(layer)
    elif activation in ["LRELU", "lrelu", "Lrelu"]:
        layer = tf.nn.leaky_relu(layer, alpha=lrelu_alpha)
    elif activation in ["SELU", "selu", "Selu"]:
        layer = tf.nn.selu(layer)
    elif activation in ["ELU", "elu", "Elu"]:
        layer = tf.nn.elu(layer)
    elif activation in ["SIGMOID", "sigmoid", "Sigmoid"]:
        layer = tf.nn.sigmoid(layer)
    elif activation in ["SOFTMAX", "softmax", "Softmax"]:
        layer == tf.nn.softmax(layer)
    
    if is_training: 
        layer = tf.nn.dropout(layer, dropout_val)
    return layer


def new_conv2d_layer(input, 
                     filter_shape, 
                     name=None, 
                     dropout_val=0.85, 
                     activation = 'LRELU', 
                     lrelu_alpha=0.2,
                     padding='SAME', 
                     strides=[1, 1, 1, 1],
                     data_type=tf.float32,  
                     is_training=True,
                     use_bias=True,
                     use_batchnorm=False):  
    """[summary]
    
    Arguments:
        input {[type]} -- [description]
        filter_shape {[type]} -- [description]
    
    Keyword Arguments:
        name {[type]} -- [description] (default: {None})
        dropout_val {float} -- [description] (default: {0.85})
        activation {str} -- [description] (default: {'LRELU'})
        lrelu_alpha {float} -- [description] (default: {0.2})
        padding {str} -- [description] (default: {'SAME'})
        strides {list} -- [description] (default: {[1, 1, 1, 1]})
        data_type {[type]} -- [description] (default: {tf.float32})
        is_training {bool} -- [description] (default: {True})
        use_bias {bool} -- [description] (default: {True})
        use_batchnorm {bool} -- [description] (default: {False})
    
    Returns:
        [type] -- [description]
    """

    shape = filter_shape
    weights = new_weights(shape=shape, name=name, data_type=data_type)
    layer = tf.nn.conv2d(input=input,
                            filter=weights,
                            strides=strides,
                            padding=padding, 
                            name='convolution_'+str(name))
    
    if use_bias:
        biases = new_biases(length=filter_shape[3], name=name, data_type=data_type)
        layer += biases
    
    if use_batchnorm:
        layer = batch_norm(inputs=layer, training=is_training)

    if activation in ["RELU", "relu", "Relu"]:
        layer = tf.nn.relu(layer)
    elif activation in ["LRELU", "lrelu", "Lrelu"]:
        layer = tf.nn.leaky_relu(layer, alpha=lrelu_alpha)
    elif activation in ["SELU", "selu", "Selu"]:
        layer = tf.nn.selu(layer)
    elif activation in ["ELU", "elu", "Elu"]:
        layer = tf.nn.elu(layer)
    elif activation in ["SIGMOID", "sigmoid", "Sigmoid"]:
        layer = tf.nn.sigmoid(layer)
    elif activation in ["SOFTMAX", "softmax", "Softmax"]:
        layer == tf.nn.softmax(layer)
    
    if is_training: 
        layer = tf.nn.dropout(layer, dropout_val)
    return layer
    

def new_conv2d_depthwise_layer(input, 
                               filter_shape, 
                               name=None, 
                               dropout_val=0.85, 
                               activation = 'LRELU', 
                               lrelu_alpha=0.2, 
                               padding='SAME', 
                               strides=[1, 1, 1, 1], 
                               data_type=tf.float32,  
                               is_training=True,
                               use_bias=True,
                               use_batchnorm=False): 
    """[summary]
    
    Arguments:
        input {[type]} -- [description]
        filter_shape {[type]} -- [description]
    
    Keyword Arguments:
        name {[type]} -- [description] (default: {None})
        dropout_val {float} -- [description] (default: {0.85})
        activation {str} -- [description] (default: {'LRELU'})
        lrelu_alpha {float} -- [description] (default: {0.2})
        padding {str} -- [description] (default: {'SAME'})
        strides {list} -- [description] (default: {[1, 1, 1, 1]})
        data_type {[type]} -- [description] (default: {tf.float32})
        is_training {bool} -- [description] (default: {True})
        use_bias {bool} -- [description] (default: {True})
        use_batchnorm {bool} -- [description] (default: {False})
    
    Returns:
        [type] -- [description]
    """
    shape = filter_shape
    weights = new_weights(shape=shape, name=name, data_type=data_type)
    layer = tf.nn.depthwise_conv2d(input=input,
                            filter=weights,
                            strides=strides,
                            padding=padding, name='convolution_depthwise_'+str(name))

    if use_bias:
        biases = new_biases(length=filter_shape[3], name=name, data_type=data_type)
        layer += biases

    if use_batchnorm:
        layer = batch_norm(inputs=layer, training=is_training)

    if activation in ["RELU", "relu", "Relu"]:
        layer = tf.nn.relu(layer)
    elif activation in ["LRELU", "lrelu", "Lrelu"]:
        layer = tf.nn.leaky_relu(layer, alpha=lrelu_alpha)
    elif activation in ["SELU", "selu", "Selu"]:
        layer = tf.nn.selu(layer)
    elif activation in ["ELU", "elu", "Elu"]:
        layer = tf.nn.elu(layer)
    elif activation in ["SIGMOID", "sigmoid", "Sigmoid"]:
        layer = tf.nn.sigmoid(layer)
    elif activation in ["SOFTMAX", "softmax", "Softmax"]:
        layer == tf.nn.softmax(layer)

    if is_training: 
        layer = tf.nn.dropout(layer, dropout_val)
    return layer
    

def new_deconv_layer(input, 
                     filter_shape, 
                     output_shape, 
                     name=None, 
                     activation = 'LRELU',  
                     lrelu_alpha=0.2, 
                     padding = 'SAME',
                     strides = [1,1,1,1],
                     data_type=tf.float32,  
                     use_bias=True):
    """[summary]
    
    Arguments:
        input {[type]} -- [description]
        filter_shape {[type]} -- [description]
        output_shape {[type]} -- [description]
    
    Keyword Arguments:
        name {[type]} -- [description] (default: {None})
        activation {str} -- [description] (default: {'LRELU'})
        lrelu_alpha {float} -- [description] (default: {0.2})
        padding {str} -- [description] (default: {'SAME'})
        strides {list} -- [description] (default: {[1,1,1,1]})
        data_type {[type]} -- [description] (default: {tf.float32})
        use_bias {bool} -- [description] (default: {True})
    
    Returns:
        [type] -- [description]
    """
    weights = tf.Variable(tf.truncated_normal(shape=[filter_shape[0], filter_shape[1], filter_shape[3], filter_shape[2]], stddev=0.05), 
                          name="weight_" + name,
                          dtype=data_type)
    deconv_shape = tf.stack(output_shape)
    layer = tf.nn.conv2d_transpose(value=input, 
                                    filter = weights, 
                                    output_shape = deconv_shape,
                                    strides = strides,
                                    padding = padding, 
                                    name="deconv_"+str(name))

    if use_bias:
        biases = new_biases(length=filter_shape[3], name=name, data_type=data_type)
        layer += biases

    if activation in ["RELU", "relu", "Relu"]:
        layer = tf.nn.relu(layer)
    elif activation in ["LRELU", "lrelu", "Lrelu"]:
        layer = tf.nn.leaky_relu(layer, alpha=lrelu_alpha)
    elif activation in ["SELU", "selu", "Selu"]:
        layer = tf.nn.selu(layer)
    elif activation in ["ELU", "elu", "Elu"]:
        layer = tf.nn.elu(layer)
    elif activation in ["SIGMOID", "sigmoid", "Sigmoid"]:
        layer = tf.nn.sigmoid(layer)
    elif activation in ["SOFTMAX", "softmax", "Softmax"]:
        layer == tf.nn.softmax(layer)

    return layer


def new_batch_norm(x, 
                   axis, 
                   phase_train, 
                   name='bn'):
    """
    Batch normalization on convolutional maps.
    Args:
        x:           Tensor, 4D BHWD input maps
        phase_train: boolean tf.Varialbe, true indicates training phase
        name:       string, variable scope
    Return:
        normed:      batch-normalized maps
    """
    
    beta = tf.Variable(tf.constant(0.0, shape=[x.get_shape().as_list()[-1]]), name='beta_' + name)
    gamma = tf.Variable(tf.constant(1.0, shape=[x.get_shape().as_list()[-1]]), name='gamma_' + name)
    mean, var = tf.nn.moments(x, axis, name='moments_' + name)
    '''
    ema = tf.train.ExponentialMovingAverage(decay=0.5)

    def mean_var_with_update():
        ema_apply_op = ema.apply([batch_mean, batch_var])
        with tf.control_dependencies([ema_apply_op]):
            return tf.identity(batch_mean), tf.identity(batch_var)

    mean, var = tf.cond(tf.cast(phase_train, tf.bool),
                        mean_var_with_update,
                        lambda: (ema.average(batch_mean), ema.average(batch_var)))
    '''
    normed = tf.nn.batch_normalization(x, mean, var, beta, gamma, 1e-3)
    
    return normed, beta, gamma


def batch_norm(inputs, 
               training, 
               momentum=0.9,
               epsilon=1e-5,
               scale=True,
               data_format='channel_last'):
    """Performs a batch normalization using a standard set of parameters.
    
    Arguments:
        inputs {[type]} -- [description]
        training {[type]} -- [description]
    
    Keyword Arguments:
        data_format {str} -- [description] (default: {'channel_last'})
    
    Returns:
        [type] -- [description]
    """
    """
    return tf.layers.batch_normalization(
        inputs=inputs, axis=1 if data_format == 'channels_first' else 3,
        momentum=momentum, epsilon=epsilon,
        scale=scale, training=training)
    """
    
    bn = tf.keras.layers.BatchNormalization(axis=1 if data_format=='channels_first' else 3, 
                                                momentum=momentum, 
                                                epsilon=epsilon, 
                                                center=True, 
                                                scale=True,
                                                renorm=False,
                                                trainable=training)
    return bn(inputs)






import tensorflow as tf
from face_recog.net import get_resnet


#images = tf.placeholder(name='img_inputs', shape=[None, 112, 112, 3], dtype=tf.float32)
#dropout_rate = tf.placeholder(name='dropout_rate', dtype=tf.float32)
w_init_method = tf.contrib.layers.xavier_initializer(uniform=False)

image_height, image_width = 112, 112
input_string = tf.placeholder(tf.string, shape=[None], name='string_input')
decode = lambda raw_byte_str: tf.image.resize_images(
                                        tf.cast(
                                            tf.image.decode_jpeg(raw_byte_str, channels=3, name='decoded_image'),
                                                tf.uint8), 
                                        [image_height, image_width])
input_images = tf.map_fn(decode, input_string, dtype=tf.float32) - 127.5
input_images = input_images / 127.5

net = get_resnet(input_images, 50, type='ir', w_init=w_init_method, trainable=False)
embedding_tensor = net.outputs

saver = tf.train.Saver()
session = tf.Session()
saver.restore(session, "/home/model/facerecog/ckpt_model_d/InsightFace_iter_best_710000.ckpt")
print ("===============")


############  serving model procedure #################
builder = tf.saved_model.builder.SavedModelBuilder('/home/model/facerecog/ckpt_model_d/serving2/')

# Create aliase tensors
# tensor_info_x: for input tensor
# tensor_info_y: for output tensor
tensor_info_x = tf.saved_model.utils.build_tensor_info(input_string)
tensor_info_y = tf.saved_model.utils.build_tensor_info(embedding_tensor)

# create prediction signature
prediction_signature = tf.saved_model.signature_def_utils.build_signature_def(
        inputs={'input': tensor_info_x},
        outputs={'output': tensor_info_y},
        method_name=tf.saved_model.signature_constants.PREDICT_METHOD_NAME)

# build frozen graph
legacy_init_op = tf.group(tf.tables_initializer(), name='legacy_init_op')
builder.add_meta_graph_and_variables(
    session, [tf.saved_model.tag_constants.SERVING],
    signature_def_map={
        'serving_default':
        prediction_signature
    },
    legacy_init_op=legacy_init_op)
builder.save()



import json
import cv2
import math
import random
import numpy as np
import tensorflow as tf
import qoalai.networks.densenet as densenet
from simple_tensor.tensor_operations import *
from simple_tensor.tensor_metrics import calculate_acc
from simple_tensor.tensor_losses import mse_loss_mean, sigmoid_crossentropy_mean
from comdutils.file_utils import *
slim = tf.contrib.slim



class Landmark(object):
    def __init__(self,
                 num_landmark_point,
                 input_height = 300,
                 input_width = 300, 
                 input_channel = 3,
                 grid_num = 3,
                 threshold = 0.5):

        """Constructor
        
        Arguments:
            classes {list of string} -- the image classes list

        Keyword Arguments:
            input_height {int} -- the height of input image (default: {512})
            input_width {int} -- the width of input image (default: {512})
            input_channel {int} -- the channel of input image (default: {3})
        """
        self.num_landmark_point = num_landmark_point
        self.input_height = input_height
        self.input_width = input_width
        self.input_channel = input_channel
        self.grid_num = grid_num
        self.grid_size_relative = 1./self.grid_num
        self.threshold = threshold

        self.input_placeholder = tf.placeholder(tf.float32, shape=(None, self.input_height, self.input_width, self.input_channel))
        self.point_position_placeholder = tf.placeholder(tf.float32, shape=(None, 1, self.num_landmark_point, 2))
        self.point_map_placeholder = tf.placeholder(tf.float32, shape=(None, self.grid_num, self.grid_num, num_landmark_point))

    
    def build_densenet_base(self, input_tensor,
                            dropout_rate,
                            is_training,
                            top_layer_depth):
        """[summary]
        
        Arguments:
            input_tensordropout_rate {[type]} -- [description]
            is_training {bool} -- [description]
        """
        # ----------------------------------- #
        # building the densenet base network  #
        # ----------------------------------- #
        arg_scoope = densenet.densenet_arg_scope()
        out = None

        with slim.arg_scope(arg_scoope):
            out = densenet.densenet121(inputs=input_tensor, 
                                       num_classes=1, 
                                       is_training=is_training)
            base_var_list = tf.get_collection(tf.GraphKeys.GLOBAL_VARIABLES)

        out = new_conv2d_layer(input=out, 
                               filter_shape=[2, 2, out.get_shape().as_list()[-1], top_layer_depth], 
                               dropout_val=1.-dropout_rate, 
                               activation = 'NONE', 
                               padding='SAME', 
                               strides=[1, 2, 2, 1],
                               is_training=is_training,
                               use_bias=False,
                               use_batchnorm=False) 
        
        # ----------------------------------- #
        # preparing the point hetamap         #
        # ----------------------------------- #
        self.point_map_nonlogit = new_conv2d_layer(input=out, 
                               filter_shape=[2, 2, out.get_shape().as_list()[-1], self.num_landmark_point], 
                               dropout_val=1-dropout_rate, 
                               activation = 'NONE', 
                               padding='SAME', 
                               strides=[1, 1, 1, 1],
                               is_training=is_training,
                               use_bias=False,
                               use_batchnorm=False)
        first_map = self.point_map_nonlogit[:, :, :, 0]
        second_map = self.point_map_nonlogit[:, :, :, 1]
        third_map = self.point_map_nonlogit[:, :, :, 2]
        fourth_map = self.point_map_nonlogit[:, :, :, 3]
        
        first_map = tf.nn.softmax(tf.keras.layers.Flatten()(first_map))
        second_map = tf.nn.softmax(tf.keras.layers.Flatten()(second_map))
        third_map = tf.nn.softmax(tf.keras.layers.Flatten()(third_map))
        fourth_map = tf.nn.softmax(tf.keras.layers.Flatten()(fourth_map))

        first_map = tf.reshape(first_map, [tf.shape(first_map)[0], self.grid_num, self.grid_num, 1])
        second_map = tf.reshape(second_map, [tf.shape(second_map)[0], self.grid_num, self.grid_num, 1])
        third_map = tf.reshape(third_map, [tf.shape(third_map)[0], self.grid_num, self.grid_num, 1])
        fourth_map = tf.reshape(fourth_map, [tf.shape(fourth_map)[0], self.grid_num, self.grid_num, 1])

        self.point_map = tf.concat([first_map, second_map, third_map, fourth_map], axis=-1)
        
        # ----------------------------------- #
        # preparing the x center predictor    #
        # preparing the y center predictor    #
        # the value is relative to the grid size
        # ----------------------------------- #
        out = tf.keras.layers.Flatten()(out)
        out_x = new_fc_layer(out, 
                             num_inputs=int(out.get_shape().as_list()[-1]), 
                             num_outputs=1000,
                             dropout_val=1.-dropout_rate, 
                             activation="NONE",
                             is_training=is_training,
                             use_bias=False)
        out_x = new_fc_layer(out_x, 
                             num_inputs=int(out_x.get_shape().as_list()[-1]), 
                             num_outputs=self.num_landmark_point,
                             dropout_val=1., 
                             activation="sigmoid",
                             is_training=is_training,
                             use_bias=False)
        
        out_y = new_fc_layer(out, 
                             num_inputs=int(out.get_shape().as_list()[-1]), 
                             num_outputs=1000,
                             dropout_val=1.-dropout_rate, 
                             activation="NONE",
                             is_training=is_training,
                             use_bias=False)
        out_y = new_fc_layer(out_y, 
                             num_inputs=int(out_y.get_shape().as_list()[-1]), 
                             num_outputs=self.num_landmark_point,
                             dropout_val=1., 
                             activation="sigmoid",
                             is_training=is_training,
                             use_bias=False)

        out_x = tf.expand_dims(out_x, 1)
        out_y = tf.expand_dims(out_y, 1)
        out_x = tf.expand_dims(out_x, -1)
        out_y = tf.expand_dims(out_y, -1)
        self.point_position = tf.nn.sigmoid(tf.concat([out_x, out_y], axis=-1))
        
        print ('--------------------INFO---------------------')
        print ('The shape of your point map: ', self.point_map)
        print ('The shape of your point position: ', self.point_position)
        print ('---------------------------------------------')


    def calculate_loss(self):
        """[summary]
        
        Returns:
            [type] -- [description]
        """
        self.point_map_loss = mse_loss_mean(self.point_map, self.point_map_placeholder)
        self.point_position_loss = mse_loss_mean(self.point_position, self.point_position_placeholder)
        self.total_loss = 2 * self.point_map_loss + self.point_position_loss
        self.point_map_acc = self.object_accuracy(self.point_map, self.point_map_placeholder)
        return self.total_loss


    def object_accuracy(self, objectness_pred, objectness_label):
        """[summary]
        
        Arguments:
            objectness_pred {[type]} -- [description]
            objectness_label {[type]} -- [description]
        
        Returns:
            [type] -- [description]
        """
        objectness_mask = tf.math.greater(objectness_pred, tf.convert_to_tensor(np.array(self.threshold), tf.float32))
        objectness_mask = tf.cast(objectness_mask, tf.float32)
        delta = (objectness_label - objectness_mask) * objectness_label
        obj_acc = 1. - tf.reduce_sum(delta) / (tf.reduce_sum(objectness_label) + 0.0001)
        return obj_acc


    def read_target(self, file_path):
        # ----------------------------------- #
        # create initial label for point map and position   
        # ----------------------------------- #
        point_position_label = np.zeros((1, self.num_landmark_point,  2))
        point_map_lable = np.zeros((self.grid_num, self.grid_num, self.num_landmark_point))

        # ----------------------------------- #
        # reading the label in a .txt file foreach image 
        #------------------------------------ #
        file = open(file_path, "r") 
        data = file.read()
        data = data.split()
        length = len(data)
        
        if length != 2 * self.num_landmark_point:
            print ("-------------------WARNING-------------------")
            print ("WRONG label format")
            print ("the label format is x1, y1, x2, y2, ..., xn, yn")
            print ("the value of x and y are relative to image width and height")
            print ("each image has single .txt file as label with the name")
            print ("---------------------------------------------")

        for i in range(self.num_landmark_point):
            # ----------------------------------- #
            # filling the point map label         #
            #------------------------------------ #
            data_x = float(data[i*2])
            data_y = float(data[i*2 + 1])
            cell_x = int(math.floor(data_x / self.grid_size_relative))
            cell_y = int(math.floor(data_y / self.grid_size_relative))
            point_map_lable[cell_y, cell_x, i] = 1.0

            # ----------------------------------- #
            # filling the point position label    #
            #------------------------------------ #
            point_position_label[0, i, 0] = (data_x - cell_x * self.grid_size_relative)/self.grid_size_relative
            point_position_label[0, i, 1] = (data_y - cell_y * self.grid_size_relative)/self.grid_size_relative

        return point_map_lable, point_position_label


    def batch_generator(self, batch_size, dataset_path, message):
        """Train Generator
        
        Arguments:
            batch_size {integer} -- the size of the batch
            image_name_list {list of string} -- the list of image name
        """
        label_folder_path = dataset_path + "labels/"
        dataset_folder_path = dataset_path + "images/"
        dataset_file_list = get_filenames(dataset_folder_path)
        random.shuffle(dataset_file_list)
        
        print ("------------------------INFO-------------------")
        print ("Image Folder: " + dataset_folder_path)
        print ("Number of Image: " + str(len(dataset_file_list)))
        print ("-----------------------------------------------")

        # Infinite loop.
        idx = 0
        while True:
            x_batch = []
            y_pred_map = []
            y_pred_position = []

            for i in range(batch_size):
                if idx >= len(dataset_file_list):
                    random.shuffle(dataset_file_list)
                    print ("==>>> INFO: your " + message +" dataset is reshuffled again: ", idx)
                    idx = 0
                    
                try:
                    tmp_x = cv2.imread(dataset_folder_path + dataset_file_list[idx])
                    tmp_x = cv2.cvtColor(tmp_x, cv2.COLOR_BGR2RGB)
                    tmp_x = cv2.resize(tmp_x, (self.input_width, self.input_height))
                    tmp_x = tmp_x.astype(np.float32) / 255.
                    tmp_y_map, tmp_y_position = self.read_target(label_folder_path + dataset_file_list[idx].split('.')[0] + ".txt")
                    x_batch.append(tmp_x)
                    y_pred_map.append(tmp_y_map)
                    y_pred_position.append(tmp_y_position)

                except Exception as e:
                    print ("---------------------------------------------------------------")
                    print ("WARNING: the " + str(e))
                    print ("---------------------------------------------------------------")

                idx += 1
            yield (np.array(x_batch), np.array(y_pred_map), np.array(y_pred_position))


    def optimize(self, 
                 iteration, 
                 subdivition,
                 cost_tensor,
                 optimizer_tensor,
                 session,
                 saver, 
                 train_generator,
                 val_generator,
                 best_loss,
                 path_tosave_model='model/model1'):
        """[summary]
        
        Arguments:
            iteration {[type]} -- [description]
            subdivition {[type]} -- [description]
            cost_tensor {[type]} -- [description]
            optimizer_tensor {[type]} -- [description]
            out_tensor {[type]} -- [description]
        
        Keyword Arguments:
            train_batch_size {int} -- [description] (default: {32})
            val_batch_size {int} -- [description] (default: {50})
            path_tosave_model {str} -- [description] (default: {'model/model1'})
        """

        self.train_loss = []
        self.val_loss = []
        best_loss = best_loss
        
        for i in range(iteration):
            sign = "-"
            t_losses = []

            for j in range(subdivition):
                x_train, y_train_map, y_train_position = next(train_generator)
                feed_dict = {}
                feed_dict[self.input_placeholder] = x_train
                feed_dict[self.point_map_placeholder] = y_train_map
                feed_dict[self.point_position_placeholder] = y_train_position
                session.run(optimizer_tensor, feed_dict)
                loss = session.run(cost_tensor, feed_dict)
                t_losses.append(loss)
                print ("> Train sub", j, 'loss : ', loss)
                
            x_val, y_val_map, y_val_position = next(val_generator)
            feed_dict = {}
            feed_dict[self.input_placeholder] = x_val
            feed_dict[self.point_map_placeholder] = y_val_map
            feed_dict[self.point_position_placeholder] = y_val_position
            loss = session.run(cost_tensor, feed_dict)
            map_acc, pml, ppl = session.run([self.point_map_acc, self.point_map_loss, self.point_position_loss], feed_dict)
            
            t_loss = sum(t_losses) / (len(t_losses) + 0.0001)
            self.train_loss.append(t_loss)
            self.val_loss.append(loss)
                
            if best_loss > loss:
                best_loss = loss
                sign = "************* model saved"
                saver.save(session, path_tosave_model)
        
            print (">> epoch: ", i, "train loss: ", round(t_loss, 3), "val loss: ", round(loss, 3), \
                "val map acc: ", round(map_acc, 3), round(pml, 3), round(ppl, 3), sign)  



# Copyright 2016 pudae. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Contains the definition of the DenseNet architecture.
As described in https://arxiv.org/abs/1608.06993.
  Densely Connected Convolutional Networks
  Gao Huang, Zhuang Liu, Kilian Q. Weinberger, Laurens van der Maaten
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf

slim = tf.contrib.slim


@slim.add_arg_scope
def _global_avg_pool2d(inputs, data_format='NHWC', scope=None, outputs_collections=None):
  with tf.variable_scope(scope, 'xx', [inputs]) as sc:
    axis = [1, 2] if data_format == 'NHWC' else [2, 3]
    net = tf.reduce_mean(inputs, axis=axis, keepdims=True)
    net = slim.utils.collect_named_outputs(outputs_collections, sc.name, net)
    return net


@slim.add_arg_scope
def _conv(inputs, num_filters, kernel_size, stride=1, dropout_rate=None,
          scope=None, outputs_collections=None):
  with tf.variable_scope(scope, 'xx', [inputs]) as sc:
    net = slim.batch_norm(inputs)
    net = tf.nn.relu(net)
    net = slim.conv2d(net, num_filters, kernel_size, stride)

    if dropout_rate:
      net = tf.nn.dropout(net)

    net = slim.utils.collect_named_outputs(outputs_collections, sc.name, net)

  return net


@slim.add_arg_scope
def _conv_block(inputs, num_filters, data_format='NHWC', scope=None, outputs_collections=None):
  with tf.variable_scope(scope, 'conv_blockx', [inputs]) as sc:
    net = inputs
    net = _conv(net, num_filters*4, 1, scope='x1')
    net = _conv(net, num_filters, 3, scope='x2')
    if data_format == 'NHWC':
      net = tf.concat([inputs, net], axis=3)
    else: # "NCHW"
      net = tf.concat([inputs, net], axis=1)

    net = slim.utils.collect_named_outputs(outputs_collections, sc.name, net)

  return net


@slim.add_arg_scope
def _dense_block(inputs, num_layers, num_filters, growth_rate,
                 grow_num_filters=True, scope=None, outputs_collections=None):

  with tf.variable_scope(scope, 'dense_blockx', [inputs]) as sc:
    net = inputs
    for i in range(num_layers):
      branch = i + 1
      net = _conv_block(net, growth_rate, scope='conv_block'+str(branch))

      if grow_num_filters:
        num_filters += growth_rate

    net = slim.utils.collect_named_outputs(outputs_collections, sc.name, net)

  return net, num_filters


@slim.add_arg_scope
def _transition_block(inputs, num_filters, compression=1.0,
                      scope=None, outputs_collections=None, code=0):

  num_filters = int(num_filters * compression)
  with tf.variable_scope(scope, 'transition_blockx', [inputs]) as sc:
    net = inputs
    if code==0: s=1
    else: s=2
    net = _conv(net, num_filters, 1, s, scope='blk')

    #net = slim.avg_pool2d(net, 1)

    net = slim.utils.collect_named_outputs(outputs_collections, sc.name, net)

  return net, num_filters


def densenet(inputs,
             num_classes=1000,
             reduction=None,
             growth_rate=None,
             num_filters=None,
             num_layers=None,
             dropout_rate=None,
             data_format='NHWC',
             is_training=True,
             reuse=None,
             scope=None):
  assert reduction is not None
  assert growth_rate is not None
  assert num_filters is not None
  assert num_layers is not None

  compression = 1.0 - reduction
  num_dense_blocks = len(num_layers)

  if data_format == 'NCHW':
    inputs = tf.transpose(inputs, [0, 3, 1, 2])

  with tf.variable_scope(scope, 'densenetxxx', [inputs, num_classes],
                         reuse=reuse) as sc:
    end_points_collection = sc.name + '_end_points'
    with slim.arg_scope([slim.batch_norm, slim.dropout],
                         is_training=is_training), \
         slim.arg_scope([slim.conv2d, _conv, _conv_block,
                         _dense_block, _transition_block], 
                         outputs_collections=end_points_collection), \
         slim.arg_scope([_conv], dropout_rate=dropout_rate):
      net = inputs

      # initial convolution
      net = slim.conv2d(net, num_filters, 7, stride=2, scope='conv1')
      net = slim.batch_norm(net)
      net = tf.nn.relu(net)
      net = slim.max_pool2d(net, 3, stride=2, padding='SAME')

      # blocks
      for i in range(num_dense_blocks - 1):
        # dense blocks
        net, num_filters = _dense_block(net, num_layers[i], num_filters,
                                        growth_rate,
                                        scope='dense_block' + str(i+1))

        # Add transition_block
        if i == 1 or i == 2 : code = 1
        else: code = 0
        net, num_filters = _transition_block(net, num_filters,
                                             compression=compression,
                                             scope='transition_block' + str(i+1), code=code)
  
      net, num_filters = _dense_block(
              net, num_layers[-1], num_filters,
              growth_rate,
              scope='dense_block' + str(num_dense_blocks))
      
      return net


def densenet121(inputs, num_classes=1000, data_format='NHWC', is_training=True, reuse=None):
  return densenet(inputs,
                  num_classes=num_classes, 
                  reduction=0.5,
                  growth_rate=32,
                  num_filters=64,
                  num_layers=[6,12,24,16],
                  data_format=data_format,
                  is_training=is_training,
                  reuse=reuse,
                  scope='densenet121')
densenet121.default_image_size = 224


def densenet161(inputs, num_classes=1000, data_format='NHWC', is_training=True, reuse=None):
  return densenet(inputs,
                  num_classes=num_classes, 
                  reduction=0.5,
                  growth_rate=48,
                  num_filters=96,
                  num_layers=[6,12,36,24],
                  data_format=data_format,
                  is_training=is_training,
                  reuse=reuse,
                  scope='densenet161')
densenet161.default_image_size = 224


def densenet169(inputs, num_classes=1000, data_format='NHWC', is_training=True, reuse=None):
  return densenet(inputs,
                  num_classes=num_classes, 
                  reduction=0.5,
                  growth_rate=32,
                  num_filters=64,
                  num_layers=[6,12,32,32],
                  data_format=data_format,
                  is_training=is_training,
                  reuse=reuse,
                  scope='densenet169')
densenet169.default_image_size = 224


def densenet_arg_scope(weight_decay=1e-4,
                       batch_norm_decay=0.99,
                       batch_norm_epsilon=1.1e-5,
                       data_format='NHWC'):
  with slim.arg_scope([slim.conv2d, slim.batch_norm, slim.avg_pool2d, slim.max_pool2d,
                       _conv_block, _global_avg_pool2d],
                      data_format=data_format):
    with slim.arg_scope([slim.conv2d],
                         weights_regularizer=slim.l2_regularizer(weight_decay),
                         activation_fn=None,
                         biases_initializer=None):
      with slim.arg_scope([slim.batch_norm],
                          scale=True,
                          decay=batch_norm_decay,
                          epsilon=batch_norm_epsilon) as scope:
        return scope



from setuptools import setup, find_packages
from qoalai import version

#from os import path
#this_directory = path.abspath(path.dirname(__file__))
#with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
#    long_description = f.read()

setup(
    name='qoalai',  
    version=version,  
    description='qoala.id data science team library',  
    long_description='',  
    long_description_content_type='text/markdown',
    author='qoala ai team',  
    packages=["qoalai", "qoalai.object_detector", "qoalai.landmarks", "qoalai.networks", "qoalai.segmentations", "qoalai.transfer_learning", "qoalai.similarity"], 
)
